package com.adlerosn.brasilfurfest.updater.startup

import android.content.Context
import android.content.Intent
import android.os.Bundle
import androidx.appcompat.app.AppCompatActivity
import android.view.Menu
import android.view.MenuItem
import android.view.View
import android.widget.ImageView
import com.adlerosn.brasilfurfest.R
import com.adlerosn.brasilfurfest.SplashActivity
import com.adlerosn.brasilfurfest.helper.asColorStateList
import com.adlerosn.brasilfurfest.helper.getColorCompat
import com.adlerosn.brasilfurfest.schedule.managers.ConventionJsonReader
import com.adlerosn.brasilfurfest.schedule.managers.RemoteAssets
import com.adlerosn.brasilfurfest.schedule.managers.UpdateChecker
import kotlinx.android.synthetic.main.activity_updater.*
import org.jetbrains.anko.opaque
import org.jetbrains.anko.runOnUiThread
import java.io.InputStream
import java.io.PrintWriter
import java.io.StringWriter


class UpdaterActivity : AppCompatActivity() {
    var updateServiceConnection: UpdateServiceConnection? = null
    var splashLaunched = false
    var currentIcon = 0
    var icons = listOf<ImageView>()
    var allowsSkipping = false

    override fun onCreate(savedInstanceState: Bundle?) {
        super.onCreate(savedInstanceState)
        if(RemoteAssets.onlineApp.not() or RemoteAssets.preloaded){
            launchSplash()
            return
        }
        setContentView(R.layout.activity_updater)
        setSupportActionBar(toolbar)
        icons = listOf(
                this@UpdaterActivity.statusAppRunning,
                this@UpdaterActivity.statusLoadFromDisk,
                this@UpdaterActivity.statusConnecting,
                this@UpdaterActivity.statusDownloading,
                this@UpdaterActivity.statusSaving,
                this@UpdaterActivity.statusListingImages,
                this@UpdaterActivity.statusDownloadingImages
        )
        updateServiceConnection = UpdateServiceConnection { baseContext.runOnUiThread { retry() } }
        baseContext.bindService(Intent(baseContext, UpdaterService::class.java), updateServiceConnection!!, Context.BIND_AUTO_CREATE)
    }

    val colorGreen = 0x43a047.opaque
    val colorYellow = 0xffb300.opaque
    val colorRed = 0xe53935.opaque
    val colorBlue = 0x039be5.opaque

    private fun setViewWorking(){
        if(currentIcon >= icons.size) currentIcon=icons.size-1
        icons[currentIcon].setImageResource(R.drawable.ic_keyboard_arrow_right_black_24dp)
        icons[currentIcon].imageTintList = colorBlue.asColorStateList
    }

    private fun setViewSuccess(){
        if(currentIcon >= icons.size) currentIcon=icons.size-1
        icons[currentIcon].setImageResource(R.drawable.ic_check_black_24dp)
        icons[currentIcon].imageTintList = colorGreen.asColorStateList
    }

    private fun setViewFailure(){
        if(currentIcon >= icons.size) currentIcon=icons.size-1
        icons[currentIcon].setImageResource(R.drawable.ic_close_black_24dp)
        icons[currentIcon].imageTintList = colorRed.asColorStateList
    }

    private fun setViewSkip(){
        if(currentIcon >= icons.size) currentIcon=icons.size-1
        icons[currentIcon].setImageResource(R.drawable.ic_redo_black_24dp)
        icons[currentIcon].imageTintList = colorYellow.asColorStateList
    }

    private fun retry() {
        val service = updateServiceConnection!!.updaterService!!
        this@UpdaterActivity.retryButton.visibility = View.INVISIBLE
        this@UpdaterActivity.errorContainer.visibility = View.GONE
        this@UpdaterActivity.errorText.text = ""
        this@UpdaterActivity.imgDownloadProgress.text = ""
        icons.forEach {
            it.setImageResource(R.drawable.ic_more_horiz_black_24dp)
            it.imageTintList = getColorCompat(R.color.colorForegroundMuted).asColorStateList
        }
        currentIcon = 0
        setViewWorking()
        setViewSuccess()
        currentIcon += 1
        if(service.hasBaseJsonLoaded()){
            setViewSuccess()
            unlockSkip()
        } else {
            setViewFailure()
        }
        currentIcon += 1
        service.updateBaseJson { downloadState: UpdateChecker.DownloadState, inputStream: ByteArray?, throwable: Throwable? ->
            when(downloadState){
                UpdateChecker.DownloadState.CONNECTING -> {
                    setViewWorking()
                }
                UpdateChecker.DownloadState.CACHE_MISS -> {
                    setViewSuccess()
                    currentIcon += 1
                    setViewSuccess()
                    currentIcon += 1
                    setViewWorking()
                }
                UpdateChecker.DownloadState.CACHE_HIT -> {
                    setViewSuccess()
                    currentIcon += 1
                    setViewSkip()
                    currentIcon += 1
                    setViewWorking()
                }
                UpdateChecker.DownloadState.ERROR -> {
                    setViewFailure()
                    throwable!!
                    val writer = StringWriter()
                    throwable.printStackTrace(PrintWriter(writer))
                    val s = writer.toString()
                    this@UpdaterActivity.retryButton.visibility = View.VISIBLE
                    this@UpdaterActivity.errorContainer.visibility = View.VISIBLE
                    this@UpdaterActivity.errorText.text = s
                    this@UpdaterActivity.retryButton.setOnClickListener { retry() }
                }
                UpdateChecker.DownloadState.SUCCESS -> {
                    setViewSuccess()
                    unlockSkip()
                    currentIcon += 1
                    checkImagesToDownload(inputStream!!.inputStream())
                }
            }
        }
    }

    private fun checkImagesToDownload(inputStream: InputStream) {
        setViewWorking()
        val allLanguageImages = ConventionJsonReader.readFromInputStream(inputStream).allImages
        val allImages = allLanguageImages.flatMap { it.values }.distinct()
        setViewSuccess()
        currentIcon += 1
        setViewWorking()
        downloadImages(allImages)
    }

    private var downloadedImages = 0
    private var totalImages = 0

    private fun incrementDownloadedImages(){
        synchronized(::downloadedImages){
            downloadedImages+=1
            this@UpdaterActivity.imgDownloadProgress.text = getString(R.string._02d_of_02d).format(downloadedImages, totalImages)
        }
        if(downloadedImages==totalImages) {
            setViewSuccess()
            updateServiceConnection!!.updaterService?.updateChecker?.cacheManager?.removeNotNeededFiles()
            launchSplash()
        }
    }

    private fun downloadImages(allImages: List<String>) {
        val context = baseContext!!
        totalImages = allImages.size
        downloadedImages = 0
        updateServiceConnection!!.updaterService?.let { service ->
            allImages.forEach {
                service.downloadImageIfNotCached(it) {
                    context.runOnUiThread {
                        incrementDownloadedImages()
                    }
                }
            }
        }
    }

    private fun unlockSkip() {
        allowsSkipping = true
        this.invalidateOptionsMenu()
    }

    private fun launchSplash(){
        if(splashLaunched) return
        splashLaunched = true
        val intent = Intent(this, SplashActivity::class.java)
        intent.flags = Intent.FLAG_ACTIVITY_NEW_TASK or Intent.FLAG_ACTIVITY_TASK_ON_HOME
        this.startActivity(intent)
        finish()
    }

    override fun onCreateOptionsMenu(menu: Menu?): Boolean {
        menuInflater.inflate(
                if(allowsSkipping)
                    R.menu.menu_skip_skip
                else
                    R.menu.menu_skip_dontskip,
                menu
        )
        return true
    }

    override fun onOptionsItemSelected(item: MenuItem): Boolean =
            when(item.itemId) {
                R.id.action_continue_in_background -> {
                    launchSplash()
                    true
                }
                else -> super.onOptionsItemSelected(item)
            }

    override fun finish() {
        if(updateServiceConnection?.connected == true) {
            updateServiceConnection?.updaterService?.stopAllDownloadSignalsFromMassDownloads()
            updateServiceConnection?.connected = false
            unbindService(updateServiceConnection!!)
        }
        super.finish()
    }
}
